#!/bin/bash

set -e

# Overwrite PATH with known safe defaults
export PATH="/usr/bin:/usr/sbin:/bin:/sbin"

appDir="$(cd "$(dirname "${BASH_SOURCE[0]}")/../.." && pwd)"
logFile="/dev/null"

readonly appName="Private Internet Access"
readonly appVersion="$(defaults read "$appDir/Contents/Info.plist" CFBundleVersion)"
readonly installDir="/Applications/$appName.app"
readonly bundleId="com.privateinternetaccess.vpn"
readonly launchDaemonPlist="/Library/LaunchDaemons/com.privateinternetaccess.vpn.daemon.plist"
readonly daemonSettingsPath="/Library/Preferences/$bundleId"
readonly oldSettingsPath="/Users/$USER/.pia_manager/data"
readonly groupName="piavpn"
readonly everyoneGUID="ABCDEFAB-CDEF-ABCD-EFAB-CDEF0000000C"

function enableLogging() {
    logFile="/Library/Application Support/com.privateinternetaccess.vpn/install.log"
    mkdir -p "$(dirname "$logFile")" || true
}
function dialog() {
    local text="$1"
    shift
    osascript > /dev/null 2> /dev/null <<EOF
set piaIcon to (POSIX file "$appDir/Contents/Resources/pia.icns") as Unicode text
display dialog "$text" with icon file piaIcon with title "$appName" $@
EOF
}
function dialogEx() {
    local extra="$(cat)"
    dialog "$@" $extra
}
function echoPass() {
    printf '\e[92m\xE2\x9C\x94\e[0m %s\n' "$@"
    echo "$(date +"[%Y-%m-%d %H:%M:%S]") [ OK ] $@" >> "$logFile" || true
}
function echoFail() {
    printf '\e[91m\xE2\x9C\x98\e[0m %s\n' "$@"
    echo "$(date +"[%Y-%m-%d %H:%M:%S]") [FAIL] $@" >> "$logFile" || true
}
function fail() {
    echoFail "$@"
    dialog "Error: $@\n\nContact support if this problem reoccurs." 'buttons {"Exit"} default button "Exit"'
    exit 1
}
function silent_pfctl() {
    local output="$(pfctl -q "$@" 2>&1)"
    local result=$?
    echo "$output" | grep -vFf <(cat <<EOF
pfctl: Use of -f option, could result in flushing of rules
present in the main ruleset added by the system at startup.
See /etc/pf.conf for further details.
EOF
    ) >> "$logFile"
    return $result
}


if [[ "$1" == "check" ]] ; then ################################################

    result=0

    # Test if we're in the global Applications directory
    if [ -d "$installDir" ] && ! [[ "$appDir/" == "$installDir/"* ]] ; then
        echoFail "Another installation found in /Applications"
        result=2
    elif [[ "$appDir" == "$installDir" ]] ; then
        echoPass "Installed in /Applications"
    elif [[ "$appDir" == "/Applications/"* ]] ; then
        echoFail "Installed in /Applications but with wrong name"
        result=2
    else
        echoFail "Not installed in /Applications"
        result=2
    fi

    # Test if the group is correctly set up
    groupID=0
    groupHasEveryone=false
    while IFS= read line ; do
        if [[ $line == "PrimaryGroupID:"* ]] ; then
            groupID="$(echo "$line" | cut -d' ' -f2-)"
        elif [[ $line == "NestedGroups:"* && $line == *"$everyoneGUID"* ]] ; then
            groupHasEveryone=true
        fi
    done <<< "$(dscl . -read "/Groups/$groupName" 2> /dev/null)"
    if [[ $groupID == 0 ]] ; then
        echoFail "Group '$groupName' does not exist"
        result=1
    else
        echoPass "Group '$groupName' exists with group ID $groupID"
    fi
    if "$groupHasEveryone" ; then
        echoPass "Group '$groupName' contains nested group 'everyone'"
    else
        echoFail "Group '$groupName' does not contain nested group 'everyone'"
        result=1
    fi

    # Check if client binary has the setgid bit and the right group ID
    #clientGroup="$(stat -f '%Sg' "$appDir/Contents/MacOS/$appName")"
    #if [[ "$clientGroup" == "$groupName" ]] ; then
    #    echoPass "Client binary has owner group '$groupName'"
    #else
    #    echoFail "Client binary has owner group '$clientGroup'"
    #    result=1
    #fi
    #if [ -g "$appDir/Contents/MacOS/$appName" ] ; then
    #    echoPass "Client binary has the sgid flag set"
    #else
    #    echoFail "Client binary does not have the sgid flag set"
    #    result=1
    #fi

    # Check if the daemon is installed
    if [ -f "$launchDaemonPlist" ] ; then
        echoPass "Daemon is registered as a LaunchDaemon"
        if grep -qsF "$appDir/Contents/MacOS/pia-daemon" <<< "$(plutil -p "$launchDaemonPlist")" ; then
            echoPass "LaunchDaemon points to the correct pia-daemon executable"
        else
            echoFail "LaunchDaemon points to a different pia-daemon executable"
            result=1
        fi
    else
        echoFail "Daemon is not registered as a LaunchDaemon"
        result=1
    fi

    exit "$result"

elif [[ "$1" == "uninstall" ]] ; then ##########################################

    [[ $EUID -eq 0 ]] || fail "Not running as root"

    enableLogging

    # Stop the daemon
    if pgrep -q pia-daemon ; then
        [ -f "$launchDaemonPlist" ] && launchctl unload "$launchDaemonPlist" 2>> "$logFile" || true
        pkill pia-daemon 2>> "$logFile" || true
        echoPass "Stopped daemon"
    fi

    # Uninstall the daemon
    if [ -f "$launchDaemonPlist" ] ; then
        rm -rf "$launchDaemonPlist" 2>> "$logFile" || echoFail "Failed to uninstall LaunchDaemon"
        echoPass "Uninstalled LaunchDaemon"
    fi

    # Stop any other processes (TODO: softer kill for clients?)
    killedProcesses=false
    while pkill -f "^$appDir/Contents/MacOS/" 2>> "$logFile" ; do
        killedProcesses=true
        sleep 0.5
    done
    if "$killedProcesses" ; then
        echoPass "Killed running processes"
    fi

    # Delete app files
    rm -rf "$appDir" 2>> "$logFile" || echoFail "Failed to remove app directory"
    echoPass "Removed app directory"

    # Clean up PF if needed
    if grep -q -F com.privateinternetaccess.vpn /etc/pf.conf ; then
        cat /etc/pf.conf | grep -v -F com.privateinternetaccess.vpn > /etc/pf.conf.new && mv -f /etc/pf.conf.new /etc/pf.conf && silent_pfctl -F rules -f /etc/pf.conf && echoPass "Uninstalled PF anchors" || echoFail "Failed to remove PF anchors"
    fi

    # Remove group
    if dscl . -read "/Groups/$groupName" > /dev/null 2> /dev/null ; then
        dscl . -delete "/Groups/$groupName" 2>> "$logFile" || true
        echoPass "Deleted group '$groupName'"
    fi

    # Remove client data
    rm -rf /Users/*/Library/Application\ Support/com.privateinternetaccess.vpn
    rm -rf /Users/*/Library/Preferences/com.privateinternetaccess.vpn
    rm -f /Users/*/Library/Preferences/com.privateinternetaccess.vpn.plist
    rm -f /Users/*/Library/LaunchAgents/com.privateinternetaccess.vpn.client.plist

    # Remove daemon data
    rm -rf /Library/Application\ Support/com.privateinternetaccess.vpn
    rm -rf /Library/Preferences/com.privateinternetaccess.vpn

    exit 0

elif [[ "$1" == "scheduleuninstall" ]] ; then ##################################

    [[ $EUID -eq 0 ]] || fail "Not running as root"

    PID="$2"

    # If a PID was passed, kill our parent and relaunch
    if [ -n "$PID" ] && [[ $PID =~ ^[0-9]+$ ]] ; then
        # Launch a disowned subshell
        (
            # Wait for parent process to die
            while kill -0 "$PID" 2> /dev/null ; do sleep 0.2 ; done
            # Execute real uninstall script
            exec "$0" uninstall
        ) & disown
    else
        fail "No PID passed"
    fi

    exit 0

else ###########################################################################

    [[ $EUID -eq 0 ]] || fail "Not running as root"

    enableLogging

    echoPass "Running from $appDir"
    [[ -z "${UNTRANSLOCATED_PATH}" ]] || echoPass "Translocated from ${UNTRANSLOCATED_PATH}"

    # Check if the old PIA client is installed
    if [ -d "$installDir" ] && grep -qF "com.privateinternetaccess.osx.PIA-VPN" "$installDir/Contents/Info.plist" ; then
        # Give the user a chance to abort
        dialogEx "This will replace your existing installation of $appName. Downgrading afterwards will require a clean reinstall. Do you wish to continue?" <<EOF || exit 0
            buttons {"Don't Continue", "Continue"} default button "Continue" cancel button "Don't Continue"
EOF
        # Block as long as the old PIA is running
        while pgrep -q -f "^/Applications/Private Internet Access.app/" ; do
            dialogEx "Please exit the existing $appName before proceeding." <<EOF || exit 0
                buttons {"Cancel", "Try Again"} default button "Try Again" cancel button "Cancel"
EOF
            sleep 1.0
        done

        # Remove old app
        rm -rf "$installDir" || fail "Failed to remove old installation"
        echoPass "Removed old installation"

        # Import settings from old app if possible
        [ "$(cd "/Users/$USER/.." && pwd)" == "/Users" ] && \
        [ -f "$oldSettingsPath/settings.json" ] && \
        [ ! -f "$daemonSettingsPath/settings.json" ] && \
        mkdir -p "$daemonSettingsPath" && \
        echo "{\"legacy\":$(cat "$oldSettingsPath/settings.json")}" > "$daemonSettingsPath/settings.json" && \
        echoPass "Migrated old settings" || \
        echoFail "Old settings not migrated"

        # Remove old app settings
        rm -rf /Users/*/.pia_manager /Library/Application\ Support/PrivateInternetAccess
    fi

    # Set up a group on the machine, which the daemon will run as in order to
    # be exempted from the killswitch and firewall rules
    if ! dscl . -read "/Groups/$groupName" > /dev/null 2>> "$logFile" ; then
        # Find an available group ID
        groupID=$(dscl . -list /Groups PrimaryGroupID 2>> "$logFile" | awk '{ g[$2]=1; } END { for (i=333;i<500;i++) { if(!g[i]) { print i; exit 0; } } exit 1; }') || fail "Failed to find available group ID"
        # Create group
        dscl . -create "/Groups/$groupName" gid "$groupID" 2>> "$logFile" || fail "Failed to create group '$groupName'"
        echoPass "Created group '$groupName' with group ID $groupID"
    fi

    # Add 'everyone' to our group as a nested group, otherwise users will not
    # be able to run binaries with the +setgid bit owned by our group on their
    # desktop (note that this does not exempt everyone's traffic)
    if ! dscl . -read "/Groups/$groupname" NestedGroups 2>> "$logFile" | grep -qF "$everyoneGUID" ; then
        dseditgroup -o edit -a everyone -t group "$groupName" || fail "Failed to add 'everyone' to group '$groupName'"
        echoPass "Added 'everyone' to group '$groupName'"
    fi

    # Help flush group changes so they're visible to other posix commands
    dscacheutil -flushcache 2>> "$logFile" || true
    id > /dev/null 2>> "$logFile" || true

    # Stop daemon if it is running
    if pgrep pia-daemon ; then
        [ -f "$launchDaemonPlist" ] && launchctl unload "$launchDaemonPlist" 2>> "$logFile" || true
        pkill pia-daemon 2>> "$logFile" || true
        echoPass "Stopped daemon"
    else
        [ -f "$launchDaemonPlist" ] && launchctl unload "$launchDaemonPlist" 2>> "$logFile" || true
    fi

    # Reload the PF rules
    silent_pfctl -f /etc/pf.conf || true

    # Quietly remove old piaX files if present
    if [ -d "/Applications/piaX.app" ] && grep -qF "com.privateinternetaccess.vpn" "/Applications/piaX.app/Contents/Info.plist" 2> /dev/null ; then
        while pkill -f "^/Applications/piaX.app/Contents/MacOS/" 2>> "$logFile" ; do
            sleep 0.5
        done
        rm -rf "/Applications/piaX.app"
        echoPass "Removed alpha version"
    fi

    # Move to /Applications if necessary
    if ! [[ "$appDir" == "$installDir" ]] ; then

        # Kill any existing processes running from the destination directory
        killedProcesses=false
        while pkill -f "^$installDir/Contents/MacOS/" 2>> "$logFile" ; do
            killedProcesses=true
            sleep 0.5
        done
        if "$killedProcesses" ; then
            echoPass "Killed running processes"
        fi

        # Delete the destination directory if it exists
        if [ -d "$installDir" ] ; then
            rm -rf "$installDir" 2>> "$logFile" || fail "Failed to delete existing installation"
            echoPass "Deleted existing installation"
        fi

        # If the destination directory isn't writable even though we're root,
        # we're probably running under app translocation
        if [ -w "$appDir" ] ; then
            mv -f "$appDir" "$installDir" 2>> "$logFile" || fail "Failed to move app to /Applications"
            echoPass "Moved app to /Applications"
        else
            # Copying seems to still work under app translocation
            cp -fpR "$appDir" "$installDir" 2>> "$logFile" || fail "Failed to copy app to /Applications"
            echoPass "Copied app to /Applications"
            # If we've been passed our own untranslocated path via the environment, remove it
            if [[ ! -z "${UNTRANSLOCATED_PATH}" ]]; then
                # To be on the safe side, only remove paths under /Users/
                if [[ "${UNTRANSLOCATED_PATH}" == "/Users/"* ]]; then
                    rm -rf "${UNTRANSLOCATED_PATH}" && echoPass "Removed original app location" || echoFail "Failed to remove ${UNTRANSLOCATED_PATH}"
                else
                    echoPass "Leaving original app location in place"
                fi
            else
                # As a fallback, use the original applescript approach
                osascript -e "set f to POSIX file \"$appDir\"" -e 'tell application "Finder" to move f to trash' 2>> "$logFile" || echoFail "Unable to remove original app location"
            fi
        fi
        # Try to coax Spotlight into finding the new app as soon as possible
        mdimport "$installDir" > /dev/null 2> /dev/null || true

        # Update appDir to new directory
        appDir="$installDir"

    fi
    
    # Unquarantine app if necessary
    #if xattr "$appDir" | grep -qxF com.apple.quarantine ; then
        "$appDir/Contents/MacOS/pia-unquarantine" "$appDir" 2>> "$logFile" || fail "Failed to unquarantine app bundle"
        echoPass "Unquarantined app bundle"
    #fi

    # Fix permissions (root owner, group+sgid)
    chown -R root:wheel "$appDir" 2>> "$logFile" || fail "Failed to set app ownership"
    #chgrp "$groupName" "$appDir/Contents/MacOS/$appName" 2>> "$logFile" || fail "Failed to set group owner on client executable"
    #chmod g+s "$appDir/Contents/MacOS/$appName" 2>> "$logFile" || fail "Failed to set the sgid bit on client executable"
    echoPass "Updated permissions"

    # Install daemon
    cat <<EOF > "$launchDaemonPlist" 2>> "$logFile" || fail "Failed to install LaunchDaemon"
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple/DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>Label</key>
    <string>com.privateinternetaccess.vpn.daemon</string>
    <key>ProgramArguments</key>
    <array>
        <string>$appDir/Contents/MacOS/pia-daemon</string>
    </array>
    <key>KeepAlive</key>
    <true/>
    <key>Umask</key>
    <integer>022</integer>
    <key>EnableTransactions</key>
    <false/>
</dict>
</plist>
EOF
    echoPass "Installed LaunchDaemon"

    # Launch daemon
    launchctl load -w "$launchDaemonPlist" 2>> "$logFile" || fail "Failed to launch daemon"
    for i in $(seq 1 25) ; do
        if pgrep -q pia-daemon ; then
            break
        fi
        sleep 0.2
    done
    if pgrep -q pia-daemon ; then
        echoPass "Launched daemon"
    else
        fail "Failed to launch daemon (timed out)"
    fi

    PID="$1"

    # If a PID was passed, kill our parent and relaunch
    if [ -n "$PID" ] && [[ $PID =~ ^[0-9]+$ ]] ; then
        # TODO: Check that it's a client binary?

        echoPass "Relaunching client"

        # Launch a disowned subshell
        (
            # Wait for parent process to die
            while kill -0 "$PID" 2> /dev/null ; do sleep 0.2 ; done
            # Launch new app as the logged in user (not root)
            open "$appDir"
        ) & disown

    else
        # Just launch directly
        open "$appDir"
    fi

    exit 0

fi #############################################################################
